from monk.tf_keras_1.transforms.imports import *
from monk.system.imports import *

from monk.tf_keras_1.transforms.transforms import transform_color_jitter
from monk.tf_keras_1.transforms.transforms import transform_random_affine
from monk.tf_keras_1.transforms.transforms import transform_random_horizontal_flip
from monk.tf_keras_1.transforms.transforms import transform_random_rotation
from monk.tf_keras_1.transforms.transforms import transform_random_vertical_flip
from monk.tf_keras_1.transforms.transforms import transform_mean_subtraction
from monk.tf_keras_1.transforms.transforms import transform_normalize



@accepts(dict, list, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transforms(system_dict, set_phases):
    '''
    Set transforms depending on the training, validation and testing phases.

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        set_phases (list): Phases in which to apply the transforms.

    Returns:
        dict: updated system dict
    '''
    transforms_test = [];
    transforms_train = [];
    transforms_val = [];
    transformations = system_dict["dataset"]["transforms"];
    normalize = False;
    for phase in set_phases:
        tsf = transformations[phase];
        if(phase=="train"):
            train_status = True;
            val_status = False;
            test_status = False;
        elif(phase=="val"):
            train_status = False;
            val_status = True;
            test_status = False;
        else:
            train_status = False;
            val_status = False;
            test_status = True;

        for i in range(len(tsf)):
            name = list(tsf[i].keys())[0]
            input_dict = tsf[i][name];
            train = train_status;
            val = val_status;
            test = test_status;

            if(name == "ColorJitter"):
                system_dict = transform_color_jitter(
                    system_dict, 
                    input_dict["brightness"], input_dict["contrast"], input_dict["saturation"], input_dict["hue"],
                    train, val, test, retrieve=True
                    );
            elif(name == "RandomAffine"):
                system_dict = transform_random_affine(
                    system_dict, 
                    input_dict["degrees"], input_dict["translate"], input_dict["scale"], input_dict["shear"], 
                    train, val, test, retrieve=True
                    );
            elif(name == "RandomHorizontalFlip"):
                system_dict = transform_random_horizontal_flip(
                    system_dict, 
                    input_dict["p"],
                    train, val, test, retrieve=True
                    );
            elif(name == "RandomVerticalFlip"):
                system_dict = transform_random_vertical_flip(
                    system_dict, 
                    input_dict["p"],
                    train, val, test, retrieve=True
                    );
            elif(name == "RandomRotation"):
                system_dict = transform_random_rotation(
                    system_dict, 
                    input_dict["degrees"], 
                    train, val, test, retrieve=True
                    );
            elif(name == "MeanSubtraction"):
                system_dict = transform_mean_subtraction(
                    system_dict,
                    input_dict["mean"], 
                    train, val, test, retrieve=True
                    );
            elif(name == "Normalize"):
                system_dict = transform_normalize(
                    system_dict,
                    input_dict["mean"], input_dict["std"],
                    train, val, test, retrieve=True
                    );


    return system_dict;





from monk.system.imports import *
from monk.system.common import read_json


@accepts(str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def print_summary(fname):
    '''
    Read a system dictionary file and print summary

    Args:
        fname (str): Path to file 

    Returns:
        None
    '''
    system_dict = read_json(fname);

    #############################################################################################################################
    print("");
    print("");
    print("Experiment Summary");
    
    print("");
    print("System");
    print("    Project Name:    {}".format(system_dict["project_name"]));
    print("    Project Dir:     {}".format(system_dict["project_dir_relative"]));
    print("    Experiment Name: {}".format(system_dict["experiment_name"]));
    print("    Experiment Dir:  {}".format(system_dict["experiment_dir_relative"]));
    print("    Library:         {}".format(system_dict["library"]));
    print("    Origin:          {}".format(system_dict["origin"]));
    print("");
    #############################################################################################################################




    #############################################################################################################################
    print("Dataset");
    print("    Status:       {}".format(system_dict["dataset"]["status"]));
    if(system_dict["dataset"]["status"]):
        print("    Dataset Type: {}".format(system_dict["dataset"]["dataset_type"]));
        print("    Train path:   {}".format(system_dict["dataset"]["train_path"]));
        print("    Val path:     {}".format(system_dict["dataset"]["val_path"]));
        print("    Test path:    {}".format(system_dict["dataset"]["test_path"]));
        print("    CSV Train:    {}".format(system_dict["dataset"]["csv_train"]));
        print("    CSV Val:      {}".format(system_dict["dataset"]["csv_val"]));
        print("    CSV Test:     {}".format(system_dict["dataset"]["csv_test"]));
        print("");


        print("Dataset Parameters:");
        print("    Input Size:   {}".format(system_dict["dataset"]["params"]["input_size"]));
        print("    Batch Size:   {}".format(system_dict["dataset"]["params"]["batch_size"]));
        print("    Shuffle:      {}".format(system_dict["dataset"]["params"]["train_shuffle"]));
        print("    Processors:   {}".format(system_dict["dataset"]["params"]["num_workers"]));
        print("    Num Classes:  {}".format(system_dict["dataset"]["params"]["num_classes"]));
        print("");
        

        print("Dataset Transforms:");
        print("    Train transforms: {}".format(system_dict["dataset"]["transforms"]["train"]));
        print("    Val transforms:   {}".format(system_dict["dataset"]["transforms"]["val"]));
        print("    Test transforms:  {}".format(system_dict["dataset"]["transforms"]["test"]));
    print("");
    #############################################################################################################################



    #############################################################################################################################
    print("Model");
    print("    Status:".format(system_dict["model"]["status"]));
    if(system_dict["model"]["status"]):
        print("    Model Name:                     {}".format(system_dict["model"]["params"]["model_name"]));
        print("    Use Gpu:                        {}".format(system_dict["model"]["params"]["use_gpu"]));
        print("    Use pretrained weights:         {}".format(system_dict["model"]["params"]["use_pretrained"]));
        print("    Base network weights freezed:   {}".format(system_dict["model"]["params"]["freeze_base_network"]));
        print("    Number of trainable parameters: {}".format(system_dict["model"]["params"]["num_params_to_update"]));
    print("")
    #############################################################################################################################



    #############################################################################################################################
    print("Hyper-Parameters");
    print("    Status: {}".format(system_dict["hyper-parameters"]["status"]));
    if(system_dict["hyper-parameters"]["status"]):
        print("    Optimizer:                {}".format(system_dict["hyper-parameters"]["optimizer"]));
        print("    Learning Rate Scheduler:  {}".format(system_dict["hyper-parameters"]["learning_rate_scheduler"]));
        print("    loss:                     {}".format(system_dict["hyper-parameters"]["loss"]));
        print("    Num epochs:               {}".format(system_dict["hyper-parameters"]["num_epochs"]));
    print("");
    #############################################################################################################################



    #############################################################################################################################
    print("");
    print("Dataset Settings");
    if("display_progress" in system_dict["training"]["settings"].keys()):
        print("    Status: {}".format(True));
        print("    Display progress:          {}".format(system_dict["training"]["settings"]["display_progress"]));
        print("    Display progress realtime: {}".format(system_dict["training"]["settings"]["display_progress_realtime"]));
        print("    Save intermediate models:  {}".format(system_dict["training"]["settings"]["save_intermediate_models"]));
        print("    Save training logs:        {}".format(system_dict["training"]["settings"]["save_training_logs"]));
        print("    Intermediate model prefix: {}".format(system_dict["training"]["settings"]["intermediate_model_prefix"]));
    else:
        print("    Status: {}".format(False));
    print("");
    #############################################################################################################################



    #############################################################################################################################
    print("");
    print("Training");
    print("    Status: {}".format(system_dict["training"]["status"]));
    if(system_dict["training"]["status"]):
        print("    Model dir:                      {}".format(system_dict["model_dir_relative"]));
        print("    Best validation accuracy:       {}".format(system_dict["training"]["outputs"]["best_val_acc"]));
        print("    Best validation accuracy epoch: {}".format(system_dict["training"]["outputs"]["best_val_acc_epoch_num"]));
        print("    Training time:                  {}".format(system_dict["training"]["outputs"]["training_time"]));
        print("    Epochs completed:               {}".format(system_dict["training"]["outputs"]["epochs_completed"]));
        print("    Max Gpu Usage:                  {}".format(system_dict["training"]["outputs"]["max_gpu_usage"]));
        print("");

        print("Training Log Files");
        print("    Train accuracy: {}".format(system_dict["training"]["outputs"]["log_train_acc_history_relative"]));
        print("    Train loss:     {}".format(system_dict["training"]["outputs"]["log_train_loss_history_relative"]));
        print("    Val accuracy:   {}".format(system_dict["training"]["outputs"]["log_val_acc_history_relative"]));
        print("    Val loss:       {}".format(system_dict["training"]["outputs"]["log_val_loss_history_relative"]));
    print("");
    #############################################################################################################################


    
    #############################################################################################################################
    print("External Evaluation");
    print("    Status: {}".format(system_dict["testing"]["status"]));
    if(system_dict["testing"]["status"]):
        print("    Evaluation Dataset path: {}".format(system_dict["dataset"]["test_path"]));
        print("    Num Images:              {}".format(system_dict["testing"]["num_images"]));
        print("    Num correct predictions: {}".format(system_dict["testing"]["num_correct_predictions"]));
        print("    Overall Accuracy:        {} %".format(system_dict["testing"]["percentage_accuracy"]));
    print("");
    #############################################################################################################################




from monk.gluon.models.imports import *
from monk.system.imports import *

@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def get_layer(network_layer):
    '''
    Get layer for appending it to transfer learning base model

    Args:
        network_layer (dict): Dictionary conatining all params relating to the layer

    Returns:
        neural network layer
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];
    num_ftrs = 0;
    if(layer_name == "linear"):
        layer = nn.Dense(layer_params["out_features"], weight_initializer=init.Xavier());
        num_ftrs = layer_params["out_features"];

    elif(layer_name == "dropout"):
        layer = nn.Dropout(layer_params["p"]);
    
    elif(layer_name == "relu"):
        layer = nn.Activation('relu');

    elif(layer_name == "sigmoid"):
        layer = nn.Activation('sigmoid');

    elif(layer_name == "tanh"):
        layer = nn.Activation('tanh');

    elif(layer_name == "softplus"):
        layer = nn.Activation('softrelu');

    elif(layer_name == "leakyrelu"):
        layer = nn.LeakyReLU(alpha=layer_params["negative_slope"]);

    elif(layer_name == "prelu"):
        layer = nn.PReLU(alpha_initializer=init.Xavier());

    elif(layer_name == "elu"):
        layer = nn.ELU(alpha=layer_params["alpha"]);

    elif(layer_name == "selu"):
        layer = nn.SELU();

    elif(layer_name == "swish"):
        layer = nn.Swish(beta=layer_params["beta"]);

    return layer


@accepts(dict, probability=float, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_dropout(system_dict, probability=0.5, final_layer=False):
    '''
    Append dropout layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        probability (float): Droping probability of neurons in next layer
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "dropout";
    tmp["params"] = {};
    tmp["params"]["p"] = probability;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, num_neurons=int, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_linear(system_dict, num_neurons=512, final_layer=False):
    '''
    Append dense (fully connected) layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        num_neurons (int): Number of neurons in the dense layer
        final_layer (bool): If True, then number of neurons are directly set as number of classes in dataset for single label type classification

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "linear";
    tmp["params"] = {};
    tmp["params"]["out_features"] = num_neurons;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;





@accepts(dict, alpha=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_elu(system_dict, alpha=1.0, final_layer=False):
    '''
    Append exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        alpha (float): Multiplicatve factor.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "elu";
    tmp["params"] = {};
    tmp["params"]["alpha"] = alpha;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;

@accepts(dict, negative_slope=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_leakyrelu(system_dict, negative_slope=0.01, final_layer=False):
    '''
    Append Leaky - ReLU activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        negative_slope (float): Multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "leakyrelu";
    tmp["params"] = {};
    tmp["params"]["negative_slope"] = negative_slope; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, num_parameters=int, init=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_prelu(system_dict, num_parameters=1, init=0.25, final_layer=False):
    '''
    Append Learnable parameterized rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        init (float): Initialization value for multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "prelu";
    tmp["params"] = {};
    tmp["params"]["num_parameters"] = num_parameters; 
    tmp["params"]["init"] = init; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_relu(system_dict, final_layer=False):
    '''
    Append rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "relu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_selu(system_dict, final_layer=False):
    '''
    Append scaled exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "selu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_sigmoid(system_dict, final_layer=False):
    '''
    Append sigmoid activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "sigmoid";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, beta=[int, float], threshold=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softplus(system_dict, beta=1, threshold=20, final_layer=False):
    '''
    Append softplus activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        beta (int): Multiplicative factor
        threshold (int): softplus (thresholded relu) limit 
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softplus";
    tmp["params"] = {};
    tmp["params"]["beta"] = beta; 
    tmp["params"]["threshold"] = threshold; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softsign(system_dict, final_layer=False):
    '''
    Append softsign activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softsign";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, beta=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_swish(system_dict, beta=1.0, final_layer=False):
    '''
    Append swish activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        beta (bool): Multiplicative factor
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "swish";
    tmp["params"] = {};
    tmp["params"]["beta"] = beta;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_tanh(system_dict, final_layer=False):
    '''
    Append tanh activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "tanh";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;
    
    return system_dict;




@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def custom_model_get_layer(network_layer):
    '''
    Get layer for custom network design

    Args:
        network_layer (dict): Network layer as dict

    Returns:
        neural network layer: Actual layer in native backend library
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];
    if(layer_name == "convolution1d"):
        return custom_model_layer_convolution1d(layer_params);
    elif(layer_name == "convolution2d"):
        return custom_model_layer_convolution2d(layer_params);
    elif(layer_name == "convolution3d"):
        return custom_model_layer_convolution3d(layer_params);

    elif(layer_name == "transposed_convolution1d"):
        return custom_model_layer_transposed_convolution1d(layer_params);
    elif(layer_name == "transposed_convolution2d"):
        return custom_model_layer_transposed_convolution2d(layer_params);
    elif(layer_name == "transposed_convolution3d"):
        return custom_model_layer_transposed_convolution3d(layer_params);

    elif(layer_name == "max_pooling1d"):
        return custom_model_layer_max_pooling1d(layer_params);
    elif(layer_name == "max_pooling2d"):
        return custom_model_layer_max_pooling2d(layer_params);
    elif(layer_name == "max_pooling3d"):
        return custom_model_layer_max_pooling3d(layer_params);

    elif(layer_name == "average_pooling1d"):
        return custom_model_layer_average_pooling1d(layer_params);
    elif(layer_name == "average_pooling2d"):
        return custom_model_layer_average_pooling2d(layer_params);
    elif(layer_name == "average_pooling3d"):
        return custom_model_layer_average_pooling3d(layer_params);

    elif(layer_name == "global_max_pooling1d"):
        return custom_model_layer_global_max_pooling1d(layer_params);
    elif(layer_name == "global_max_pooling2d"):
        return custom_model_layer_global_max_pooling2d(layer_params);
    elif(layer_name == "global_max_pooling3d"):
        return custom_model_layer_global_max_pooling3d(layer_params);

    elif(layer_name == "global_average_pooling1d"):
        return custom_model_layer_global_average_pooling1d(layer_params);
    elif(layer_name == "global_average_pooling2d"):
        return custom_model_layer_global_average_pooling2d(layer_params);
    elif(layer_name == "global_average_pooling3d"):
        return custom_model_layer_global_average_pooling3d(layer_params);

    elif(layer_name == "fully_connected"):
        return custom_model_layer_fully_connected(layer_params);
    elif(layer_name == "dropout"):
        return custom_model_layer_dropout(layer_params);
    elif(layer_name == "flatten"):
        return custom_model_layer_flatten(layer_params);
    elif(layer_name == "identity"):
        return custom_model_layer_identity(layer_params);

    elif(layer_name == "batch_normalization"):
        return custom_model_layer_batch_normalization(layer_params);
    elif(layer_name == "instance_normalization"):
        return custom_model_layer_instance_normalization(layer_params);
    elif(layer_name == "layer_normalization"):
        return custom_model_layer_layer_normalization(layer_params);

    elif(layer_name == "relu"):
        return custom_model_activation_relu(layer_params);
    elif(layer_name == "sigmoid"):
        return custom_model_activation_sigmoid(layer_params);
    elif(layer_name == "tanh"):
        return custom_model_activation_tanh(layer_params);
    elif(layer_name == "softplus"):
        return custom_model_activation_softplus(layer_params);
    elif(layer_name == "softsign"):
        return custom_model_activation_softsign(layer_params);

    elif(layer_name == "elu"):
        return custom_model_activation_elu(layer_params);
    elif(layer_name == "gelu"):
        return custom_model_activation_gelu(layer_params);
    elif(layer_name == "leaky_relu"):
        return custom_model_activation_leaky_relu(layer_params);
    elif(layer_name == "prelu"):
        return custom_model_activation_prelu(layer_params);
    elif(layer_name == "selu"):
        return custom_model_activation_selu(layer_params);
    elif(layer_name == "swish"):
        return custom_model_activation_swish(layer_params);







@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution1d(params):
    '''
    Append 1d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    layer = nn.Conv1D(params["output_channels"], 
                        params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        layout=params["layout"], 
                        activation=None, 
                        use_bias=params["use_bias"], 
                        weight_initializer=None, 
                        bias_initializer='zeros', 
                        in_channels=0);
    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution2d(params):
    '''
    Append 2d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    layer = nn.Conv2D(params["output_channels"], 
                        params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        layout=params["layout"], 
                        activation=None, 
                        use_bias=params["use_bias"], 
                        weight_initializer=None, 
                        bias_initializer='zeros', 
                        in_channels=0);
    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution3d(params):
    '''
    Append 3d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;
        
    layer = nn.Conv3D(params["output_channels"], 
                        params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        layout=params["layout"], 
                        activation=None, 
                        use_bias=params["use_bias"], 
                        weight_initializer=None, 
                        bias_initializer='zeros', 
                        in_channels=0);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution1d(params):
    '''
    Append 1d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;
    layer = nn.Conv1DTranspose(params["output_channels"], 
                                params["kernel_size"], 
                                strides=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"], 
                                layout=params["layout"], 
                                activation=None, 
                                use_bias=params["use_bias"], 
                                weight_initializer=None, 
                                bias_initializer='zeros', 
                                in_channels=0)
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution2d(params):
    '''
    Append 2d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;
    layer = nn.Conv2DTranspose(params["output_channels"], 
                                params["kernel_size"], 
                                strides=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"], 
                                layout=params["layout"], 
                                activation=None, 
                                use_bias=params["use_bias"], 
                                weight_initializer=None, 
                                bias_initializer='zeros', 
                                in_channels=0)
    return layer



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution3d(params):
    '''
    Append 3d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;
    layer = nn.Conv3DTranspose(params["output_channels"], 
                                params["kernel_size"], 
                                strides=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"], 
                                layout=params["layout"], 
                                activation=None, 
                                use_bias=params["use_bias"], 
                                weight_initializer=None, 
                                bias_initializer='zeros', 
                                in_channels=0)
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling1d(params):
    '''
    Append 1d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-max-pooling layer
    '''
    layer = nn.MaxPool1D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling2d(params):
    '''
    Append 2d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-max-pooling layer
    '''
    layer = nn.MaxPool2D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling3d(params):
    '''
    Append 3d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-max-pooling layer
    '''
    layer = nn.MaxPool3D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling1d(params):
    '''
    Append 1d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-average-pooling layer
    '''
    layer = nn.AvgPool1D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        count_include_pad=params["include_padding_in_calculation"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling2d(params):
    '''
    Append 2d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-average-pooling layer
    '''

    layer = nn.AvgPool2D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        count_include_pad=params["include_padding_in_calculation"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling3d(params):
    '''
    Append 3d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-average-pooling layer
    '''

    layer = nn.AvgPool3D(pool_size=params["kernel_size"], 
                        strides=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        count_include_pad=params["include_padding_in_calculation"],
                        layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling1d(params):
    '''
    Append 1d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-max-pooling layer
    '''
    layer = nn.GlobalMaxPool1D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling2d(params):
    '''
    Append 2d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-max-pooling layer
    '''
    layer = nn.GlobalMaxPool2D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling3d(params):
    '''
    Append 3d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-max-pooling layer
    '''
    layer = nn.GlobalMaxPool3D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling1d(params):
    '''
    Append 1d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-average-pooling layer
    '''
    layer = nn.GlobalAvgPool1D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling2d(params):
    '''
    Append 2d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-average-pooling layer
    '''
    layer = nn.GlobalAvgPool2D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling3d(params):
    '''
    Append 3d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-average-pooling layer
    '''
    layer = nn.GlobalAvgPool3D(layout=params["layout"]);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_fully_connected(params):
    '''
    Append fc (dense) to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: fc (dense) layer
    '''
    layer = nn.Dense(params["units"], 
                    activation=None, 
                    use_bias=params["use_bias"], 
                    flatten=params["flatten"], 
                    dtype='float32', 
                    weight_initializer=None, 
                    bias_initializer='zeros', 
                    in_units=0);
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_dropout(params):
    '''
    Append dropout to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: dropout layer
    '''
    layer = nn.Dropout(params["drop_probability"],
                        axes=params["axes"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_flatten(params):
    '''
    Append flatten to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: flatten layer
    '''
    layer = nn.Flatten();
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_identity(params):
    '''
    Append idenity to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: idenity layer
    '''
    layer = contrib_nn.Identity();
    return layer


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_batch_normalization(params):
    '''
    Append batchnorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural netwrk layer: batchnorm layer
    '''
    layer = nn.BatchNorm(axis=1, 
                        momentum=params["moving_average_momentum"], 
                        epsilon=params["epsilon"], 
                        center=params["use_trainable_parameters"],
                        scale=params["use_trainable_parameters"], 
                        use_global_stats=params["activate_scale_shift_operation"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_instance_normalization(params):
    '''
    Append instancenorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: instancenorm layer
    '''
    layer = nn.InstanceNorm(axis=1,
                        epsilon=params["epsilon"], 
                        center=params["use_trainable_parameters"],
                        scale=params["use_trainable_parameters"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_layer_normalization(params):
    '''
    Append layernorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: layernorm layer
    '''
    layer = nn.LayerNorm(axis=1, 
                        epsilon=params["epsilon"], 
                        center=params["use_trainable_parameters"],
                        scale=params["use_trainable_parameters"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_relu(params):
    '''
    Append relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: relu activation
    '''
    layer = nn.Activation("relu");
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_sigmoid(params):
    '''
    Append sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: sigmoid activation
    '''
    layer = nn.Activation("sigmoid");
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_tanh(params):
    '''
    Append tanh to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: tanh activation
    '''
    layer = nn.Activation("tanh");
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softplus(params):
    '''
    Append softplus to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softplus activation
    '''
    layer = nn.Activation("softrelu");
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softsign(params):
    '''
    Append softsign to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softsign activation
    '''
    layer = nn.Activation("softsign");
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_elu(params):
    '''
    Append elu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: elu activation
    '''
    layer = nn.ELU(alpha=params["alpha"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_gelu(params):
    '''
    Append gelu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: gelu activation
    '''
    layer = nn.GELU();
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_prelu(params):
    '''
    Append prelu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: prelu activation
    '''
    layer = nn.PReLU();
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_leaky_relu(params):
    '''
    Append leaky-relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: leaky-relu activation
    '''
    layer = nn.LeakyReLU(alpha=params["alpha"]);
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_selu(params):
    '''
    Append selu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: selu activation
    '''
    layer = nn.SELU();
    return layer;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_swish(params):
    '''
    Append swish to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: swish activation
    '''
    layer = nn.Swish(beta=params["beta"]);
    return layer;



class addBlock(nn.HybridBlock):
    '''
    Class for creating a custom block (layer) representing elementwise addition

    Args:
        branches (list): List of layers to merge with elementwise addition
    '''
    def __init__(self, branches, **kwargs):
        super(addBlock, self).__init__(**kwargs)
        self.child_names = [];     
        
        with self.name_scope():
            for i in range(len(branches)):
                vars(self)["body" + str(i)] = nn.HybridSequential(prefix='');
                for j in range(len(branches[i])):
                    vars(self)["body" + str(i)].add(branches[i][j]);
                self.child_names.append("body" + str(i));
        
        for i, child in enumerate(self.child_names):
            setattr(self, 'body{0}'.format(i), vars(self)["body" + str(i)])
        
            
    def hybrid_forward(self, F, x):
        for i in range(len(self.child_names)):
            if(i==0):
                y = vars(self)["body" + str(i)](x);
            else:
                y = y + vars(self)["body" + str(i)](x);
            
        return y













    





from monk.tf_keras_1.losses.imports import *
from monk.system.imports import *


@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def l1(system_dict, weight=None, batch_axis=0):
    '''
    Select L1 Loss

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "l1";
    system_dict["hyper-parameters"]["loss"]["name"] = "l1";
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;


@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def l2(system_dict, weight=1.0, batch_axis=0):
    '''
    Select L2 Loss

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "l2";
    system_dict["hyper-parameters"]["loss"]["name"] = "l2";
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;


@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, 
    axis_to_sum_over=int, label_as_categories=bool, label_smoothing=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def crossentropy(system_dict, weight=None, batch_axis=0, axis_to_sum_over=-1, 
    label_as_categories=True, label_smoothing=False):
    '''
    Select crossentropy Loss - Need to manually apply softmax

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N
        axis_to_sum_over (int): Set as -1
        label_as_categories (bool): Fixed as True
        label_smoothing (bool): If True, label smoothning is applied.

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "crossentropy";
    system_dict["hyper-parameters"]["loss"]["name"] = "crossentropy";
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["loss"]["params"]["axis_to_sum_over"] = axis_to_sum_over;
    system_dict["hyper-parameters"]["loss"]["params"]["label_as_categories"] = label_as_categories;
    system_dict["hyper-parameters"]["loss"]["params"]["label_smoothing"] = label_smoothing;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;



@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def binary_crossentropy(system_dict, weight=None, batch_axis=0):
    '''
    Select binary crossentropy Loss - Need to manually apply sigmoid

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "binarycrossentropy";
    system_dict["hyper-parameters"]["loss"]["name"] = "binarycrossentropy";
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;


@accepts(dict, log_pre_applied=bool, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, 
    axis_to_sum_over=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def kldiv(system_dict, log_pre_applied=False, weight=None, batch_axis=0, axis_to_sum_over=-1):
    '''
    Select lkdiv Loss

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N
        axis_to_sum_over (int): Set as -1
        log_pre_applied (bool): If set as False, then logarithmic function is auto applied over target variables

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "kldiv";
    system_dict["hyper-parameters"]["loss"]["name"] = "kldiv";
    system_dict["hyper-parameters"]["loss"]["params"]["log_pre_applied"] = log_pre_applied;
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["loss"]["params"]["axis_to_sum_over"] = axis_to_sum_over;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;


@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, margin=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def hinge(system_dict, weight=None, batch_axis=0, margin=1):
    '''
    Select hinge Loss

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N
        margin (float): MArgin value.

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "hinge";
    system_dict["hyper-parameters"]["loss"]["name"] = "hinge";
    system_dict["hyper-parameters"]["loss"]["params"]["margin"] = margin;
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;


@accepts(dict, weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, margin=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def squared_hinge(system_dict, weight=None, batch_axis=0, margin=1):
    '''
    Select squared hinge Loss

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        weight (float): global scalar for weight loss
        batch_axis (int): Axis representing number of elements in the batch - N
        margin (float): MArgin value.

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = "squaredhinge";
    system_dict["hyper-parameters"]["loss"]["name"] = "squaredhinge";
    system_dict["hyper-parameters"]["loss"]["params"]["margin"] = margin;
    system_dict["hyper-parameters"]["loss"]["params"]["weight"] = weight;
    system_dict["hyper-parameters"]["loss"]["params"]["batch_axis"] = batch_axis;
    system_dict["hyper-parameters"]["status"] = True;
    return system_dict;



from monk.gluon.optimizers.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_optimizer(system_dict):
    '''
    Retrieve Optimizers in validation, testing, copy-from and resume modes

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = system_dict["hyper-parameters"]["optimizer"]["name"];
    return system_dict;


from monk.gluon.finetune.imports import *
from monk.system.imports import *

from monk.gluon.finetune.level_4_evaluation_base import finetune_evaluation




class finetune_state(finetune_evaluation):
    '''
    Base class for Monk states - train, eval_infer, resume, copy_from, pseudo_copy_from (for running sub-experiments)

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    
    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_state_eval_infer(self):
        '''
        Set system for eval_infer state

        Args:
            None

        Returns:
            None
        '''
        self.system_dict = read_json(self.system_dict["fname"], verbose=self.system_dict["verbose"]);
        self.system_dict["states"]["eval_infer"] = True;

        if(self.system_dict["training"]["status"]):
            if(len(self.system_dict["dataset"]["transforms"]["test"])):
                self.system_dict = retrieve_test_transforms(self.system_dict);
            else:
                self.custom_print("Test transforms not found.");
                self.custom_print("Add test transforms");
                self.custom_print("");
            self.set_model_final();
        else:
            msg = "Model in {} not trained. Cannot perform testing or inferencing".format(self.system_dict["experiment_name"]);
            raise ConstraintError(msg);
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_state_resume_train(self):
        '''
        Set system for resume training state

        Args:
            None

        Returns:
            None
        '''
        self.system_dict = read_json(self.system_dict["fname"], verbose=self.system_dict["verbose"]);
        self.system_dict["states"]["resume_train"] = True;
        if(self.system_dict["dataset"]["status"]):
            self.system_dict = retrieve_trainval_transforms(self.system_dict);
            self.set_dataset_final();
        else:
            msg = "Dataset not set.\n";
            msg += "Training not started. Cannot Run resume Mode";
            raise ConstraintError(msg);
        if(self.system_dict["model"]["status"]):
            self.set_model_final();
        else:
            msg = "Model not set.\n";
            msg += "Training not started. Cannot Run resume Mode";
            raise ConstraintError(msg);
        if(self.system_dict["hyper-parameters"]["status"]):
            self.system_dict = retrieve_optimizer(self.system_dict);
            self.system_dict = retrieve_scheduler(self.system_dict);
            self.system_dict = retrieve_loss(self.system_dict);
        else:
            msg = "hyper-parameters not set.\n";
            msg += "Training not started. Cannot Run resume Mode";
            raise ConstraintError(msg);
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", list, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_state_copy_from(self, copy_from):
        '''
        Set system for copied state

        Args:
            None

        Returns:
            None
        '''
        fname = self.system_dict["master_systems_dir_relative"] + copy_from[0] + "/" + copy_from[1] + "/experiment_state.json";
        system_dict_tmp = read_json(fname, verbose=self.system_dict["verbose"]);

        if(not system_dict_tmp["training"]["status"]):
            self.custom_print("Project - {}, Experiment - {}, has incomplete training.".format(copy_from[0], copy_from[1]));
            self.custom_print("Complete Previous training before copying from it.");
            self.custom_print("");
        elif(copy_from[0] == self.system_dict["project_name"] and copy_from[1] == self.system_dict["experiment_name"]):
            self.custom_print("Cannot copy same experiment. Use a different experiment to copy and load a previous experiment");
            self.custom_print("");
        else:
            self.system_dict["dataset"] = system_dict_tmp["dataset"];
            self.system_dict["model"] = system_dict_tmp["model"];
            self.system_dict["hyper-parameters"] = system_dict_tmp["hyper-parameters"];
            self.system_dict["training"] = system_dict_tmp["training"];
            self.system_dict["origin"] = [copy_from[0], copy_from[1]];
            self.system_dict["training"]["outputs"] = {};
            self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;
            self.system_dict["training"]["outputs"]["best_val_acc"] = 0;
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = 0;
            self.system_dict["training"]["outputs"]["epochs_completed"] = 0;
            self.system_dict["training"]["status"] = False;
            self.system_dict["training"]["enabled"] = True;
            self.system_dict["testing"] = {};
            self.system_dict["testing"]["status"] = False;
            save(self.system_dict);

            self.system_dict = read_json(self.system_dict["fname_relative"], verbose=self.system_dict["verbose"]);
            self.system_dict["states"]["copy_from"] = True;
            self.system_dict = retrieve_trainval_transforms(self.system_dict);
            self.Dataset();
            self.set_model_final();
            self.system_dict = retrieve_optimizer(self.system_dict);
            self.system_dict = retrieve_scheduler(self.system_dict);
            self.system_dict = retrieve_loss(self.system_dict);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", list, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_state_pseudo_copy_from(self, pseudo_copy_from):
        '''
        Set system for copied state in hyper-parameter analysis mode 

        Args:
            None

        Returns:
            None
        '''
        fname = self.system_dict["master_systems_dir_relative"] + pseudo_copy_from[0] + "/" + pseudo_copy_from[1] + "/experiment_state.json";
        system_dict_tmp = read_json(fname, verbose=self.system_dict["verbose"]);
        self.system_dict["dataset"] = system_dict_tmp["dataset"];
        self.system_dict["model"] = system_dict_tmp["model"];
        self.system_dict["hyper-parameters"] = system_dict_tmp["hyper-parameters"];
        self.system_dict["training"] = system_dict_tmp["training"];
        self.system_dict["origin"] = [pseudo_copy_from[0], pseudo_copy_from[1]];
        self.system_dict["training"]["outputs"] = {};
        self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;
        self.system_dict["training"]["outputs"]["best_val_acc"] = 0;
        self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = 0;
        self.system_dict["training"]["outputs"]["epochs_completed"] = 0;
        self.system_dict["training"]["status"] = False;
        self.system_dict["training"]["enabled"] = True;
        self.system_dict["testing"] = {};
        self.system_dict["testing"]["status"] = False;
        save(self.system_dict);

        self.system_dict = read_json(self.system_dict["fname_relative"], verbose=self.system_dict["verbose"]);
        self.system_dict["states"]["pseudo_copy_from"] = True;
        self.system_dict = retrieve_trainval_transforms(self.system_dict);
        self.Dataset();
        self.set_model_final();
        self.system_dict = retrieve_optimizer(self.system_dict);
        self.system_dict = retrieve_scheduler(self.system_dict);
        self.system_dict = retrieve_loss(self.system_dict);
    ###############################################################################################################################################

    


from monk.pytorch.training.imports import *
from monk.system.imports import *


@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_num_epochs(num_epochs, system_dict):
    '''
    Set number of training epochs

    Args:
        num_epochs (int): Number of epochs to train the network
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["hyper-parameters"]["num_epochs"] = num_epochs;
    return system_dict;



@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_display_progress_realtime(value, system_dict):
    '''
    Set verbosity levels for iterations

    Args:
        value (bool): If True, displays progress for every iteration in the epoch
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["training"]["settings"]["display_progress_realtime"] = value;
    return system_dict;


@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_display_progress(value, system_dict):
    '''
    Set all training params for epochs

    Args:
        value (bool): If True, displays summary post every epoch
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["training"]["settings"]["display_progress"] = value;
    return system_dict;


@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_save_intermediate_models(value, system_dict):
    '''
    Set whether to save models post every epoch or not

    Args:
        value (bool): If True, saves model weight post every epoch
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["training"]["settings"]["save_intermediate_models"] = value;
    return system_dict;


@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_save_training_logs(value, system_dict):
    '''
    Set whether to save training logs or not

    Args:
        value (bool): If True, saves all training and validation metrics. Required for comparison.
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["training"]["settings"]["save_training_logs"] = value;
    return system_dict;


@accepts(str, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_intermediate_model_prefix(value, system_dict):
    '''
    Set a prefix to names of intermediate models being saved

    Args:
        value (str): Appends a prefix to intermediate weights
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["training"]["settings"]["intermediate_model_prefix"] = value;
    return system_dict;


from monk.pytorch.finetune.imports import *
from monk.system.imports import *

from monk.pytorch.finetune.level_8_layers_main import prototype_layers


class prototype_transforms(prototype_layers):
    '''
    Main class for all transforms in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @warning_checks(None, ["gte", 32, "lte", 1024], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gt", 0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", int, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_center_crop(self, input_size, train=False, val=False, test=False):
        '''
        Apply Center Cropping transformation

        Args:
            input_size (int, list): Crop size
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_center_crop(self.system_dict, input_size, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @error_checks(None, brightness=["gte", 0.0], contrast=["gte", 0.0], saturation=["gte", 0.0], hue=["gte", 0.0], 
        train=None, val=None, test=None, post_trace=False)
    @accepts("self", brightness=[int, float], contrast=[int, float], saturation=[int, float], hue=[int, float], 
        train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_color_jitter(self, brightness=0, contrast=0, saturation=0, hue=0, train=False, val=False, test=False):
        '''
        Apply Color jittering transformations

        Args:
            brightness (float): Levels to jitter brightness.
                                        0 - min
                                        1 - max
            contrast (float): Levels to jitter contrast.
                                        0 - min
                                        1 - max
            saturation (float): Levels to jitter saturation.
                                        0 - min
                                        1 - max
            hue (float): Levels to jitter hue.
                                        0 - min
                                        1 - max
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_color_jitter(self.system_dict, brightness, contrast, saturation, hue, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, ["lt", 30], translate=["lt", 1.0], scale=["lt", 3.0], shear=None, 
        train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gte", 0.0], translate=["gte", 0.0], scale=["gte", 0.0], shear=["gte", 0.0], 
        train=None, val=None, test=None, post_trace=False)
    @accepts("self", [list, float, int], translate=[tuple, type(None)], scale=[tuple, type(None)], 
        shear=[list, float, int, tuple, type(None)], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_affine(self, degrees, translate=None, scale=None, shear=None, train=False, val=False, test=False):
        '''
        Apply random affine transformations

        Args:
            degrees (float): Max Rotation range limit for transforms
            scale (float, list): Range for randomly scaling 
            shear (float, list): Range for randomly applying sheer changes
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_affine(self.system_dict, degrees, translate, scale, shear, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, ["gte", 32, "lte", 1024], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gt", 0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", int, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_crop(self, input_size, train=False, val=False, test=False):  
        '''
        Apply Random Cropping transformation

        Args:
            input_size (int, list): Crop size
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''  
        self.system_dict = transform_random_crop(self.system_dict, input_size, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @error_checks(None, probability=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", probability=float, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_horizontal_flip(self, probability=0.5, train=False, val=False, test=False):
        '''
        Apply random horizontal flip transformations

        Args:
            probability (float): Probability of flipping the input image
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_horizontal_flip(self.system_dict, probability, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @error_checks(None, distortion_scale=["gt", 0, "lt", 1], probability=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", distortion_scale=float, probability=float, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_perspective(self, distortion_scale=0.5, probability=0.5, train=False, val=False, test=False):
        '''
        Apply random perspective transformations

        Args:
            distortion_scale (float): Max limit for perspective distortion
            probability (float): Probability of applying transformation
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_perspective(self.system_dict, distortion_scale, probability, train, val, test);
    ###############################################################################################################################################





    ###############################################################################################################################################
    @warning_checks(None, ["gte", 32, "lte", 1024], scale=None, ratio=["lt", 2.5], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gt", 0], scale=["gt", 0, "lt", 1], ratio=["gt", 0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", int, scale=[tuple, float], ratio=[tuple, float], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_resized_crop(self, input_size, scale=(0.08, 1.0), ratio=(0.75, 1.3333333333333333), train=False, val=False, test=False):
        '''
        Apply Random Resized Cropping transformation

        Args:
            input_size (int, list): Crop size
            scale (float, tuple): scaling ratio limits; for maximum and minimum random scaling
            ratio (float, tuple): aspect ratio limits; for maximum and minmum changes to aspect ratios 
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_resized_crop(self.system_dict, input_size, scale, ratio, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @error_checks(None, num_output_channels=["in", [1, 3]], train=None, val=None, test=None, post_trace=False)
    @accepts("self", int, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_grayscale(self, num_output_channels=3, train=False, val=False, test=False):
        '''
        Not active
        '''
        self.system_dict = transform_grayscale(self.system_dict, num_output_channels, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, ["lt", 30], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gte", 0.0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", [float, int, list], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_rotation(self, degrees, train=False, val=False, test=False):
        '''
        Apply random rotation transformations

        Args:
            degrees (float): Max Rotation range limit for transforms
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_rotation(self.system_dict, degrees, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @error_checks(None, probability=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", probability=float, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_vertical_flip(self, probability=0.5, train=False, val=False, test=False):
        '''
        Apply random vertical flip transformations

        Args:
            probability (float): Probability of flipping the input image
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_vertical_flip(self.system_dict, probability, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @warning_checks(None, ["gte", 32, "lte", 1024], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gt", 0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", int, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_resize(self, input_size, train=False, val=False, test=False):
        '''
        Apply standard resizing

        Args:
            input_size (int, list): expected final size
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_resize_gluon(self.system_dict, input_size, train, val, test);
    ###############################################################################################################################################


    ###############################################################################################################################################
    @error_checks(None, mean=["gt", 0, "lt", 1], std=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", mean=[list, float], std=[list, float], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_normalize(self, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225], train=False, val=False, test=False):
        '''
        Apply mean subtraction and standard normalization

        Args:
            mean (float, list): Mean value for subtraction
            std (float, list): Normalization factor
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_normalize(self.system_dict, mean, std, train, val, test);
    ###############################################################################################################################################






from monk.gluon.models.imports import *
from monk.system.imports import *

#@TraceFunction(trace_args=False, trace_rv=False)
def initialize_network(net, initializer):
    '''
    Get the right initializer for custom network weight initialization

    Args:
        initializer (str): The type of initializer

    Returns:
        gluon mxnet initializer: The type of initializer
    '''
    if(initializer == "xavier_normal"):
        net.initialize(initializer_xavier_normal(), ctx = ctx);
    elif(initializer == "xavier_uniform"):
        net.initialize(initializer_xavier_uniform(), ctx = ctx);
    elif(initializer == "msra"):
        net.initialize(initializer_msra(), ctx = ctx);
    elif(initializer == "normal"):
        net.initialize(initializer_normal(), ctx = ctx);
    elif(initializer == "orthogonal_normal"):
        net.initialize(initializer_orthogonal_normal(), ctx = ctx);
    elif(initializer == "orthogonal_uniform"):
        net.initialize(initializer_orthogonal_uniform(), ctx = ctx);
    elif(initializer == "uniform"):
        net.initialize(initializer_uniform(), ctx = ctx);

    return net;
    


#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_xavier_normal():
    '''
    Get Xavier normal initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: Xavier Normal intializer
    '''
    return initializer.Xavier(rnd_type='gaussian');



#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_xavier_uniform():
    '''
    Get Xavier uniform initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: Xavier uniform intializer
    '''
    return initializer.Xavier(rnd_type='uniform');



#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_msra():
    '''
    Get MSRA initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: MSRA intializer
    '''
    return initializer.MSRAPrelu();



#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_normal():
    '''
    Get normal initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: normal intializer
    '''
    return initializer.Normal();



#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_orthogonal_normal():
    '''
    Get orthogonal normal initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: Orthogonal normal intializer
    '''
    return initializer.Orthogonal(rand_type='normal');




#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_orthogonal_uniform():
    '''
    Get orthogonal uniform initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: Orthogonal uniform intializer
    '''
    return initializer.Orthogonal(rand_type='uniform');




#@TraceFunction(trace_args=False, trace_rv=False)
def initializer_uniform():
    '''
    Get uniform initialization

    Args:
        None

    Returns:
        gluon mxnet initializer: Uniform intializer
    '''
    return initializer.Uniform();


import os
import sys

import psutil

from monk.keras_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.keras.common import print_start
from monk.pip_unit_tests.keras.common import print_status

import tensorflow as tf
if(tf.__version__[0] == '2'):
    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()
import numpy as np



def test_initializer_orthogonal(system_dict):
    forward = True;

    test = "test_initializer_orthogonal";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");


            network = [];
            network.append(gtf.convolution(output_channels=16));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.convolution(output_channels=32));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.average_pooling(kernel_size=2));


            network.append(gtf.convolution(output_channels=64));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.convolution(output_channels=64));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.average_pooling(kernel_size=2));


            network.append(gtf.convolution(output_channels=128));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.convolution(output_channels=128));
            network.append(gtf.batch_normalization());
            network.append(gtf.relu());
            network.append(gtf.average_pooling(kernel_size=2));

            network.append(gtf.flatten());
            network.append(gtf.dropout(drop_probability=0.2));
            network.append(gtf.fully_connected(units=1024));
            network.append(gtf.dropout(drop_probability=0.2));
            network.append(gtf.fully_connected(units=2));
            network.append(gtf.softmax());

            gtf.Compile_Network(network, data_shape=(3, 32, 32), network_initializer="orthogonal");


            x = tf.placeholder(tf.float32, shape=(1, 32, 32, 3))
            y = gtf.system_dict["local"]["model"](x);          

            system_dict["successful_tests"] += 1;
            print_status("Pass");

        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.tf_keras_1.transforms.imports import *
from monk.system.imports import *



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transform_estimate(system_dict):
    '''
    Load training and validation transforms in main state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''

    if(system_dict["local"]["transforms_train"]["featurewise_center"]):
        rescale_train = 1/256;
    else:
        rescale_train = 0;
    if(system_dict["local"]["transforms_val"]["featurewise_center"]):
        rescale_val = 1/256;
    else:
        rescale_val = 0;


    system_dict["local"]["data_generators"]["estimate"] = keras.preprocessing.image.ImageDataGenerator(
                            featurewise_center=system_dict["local"]["transforms_train"]["featurewise_center"],
                            featurewise_std_normalization=system_dict["local"]["transforms_train"]["featurewise_std_normalization"],
                            rotation_range=system_dict["local"]["transforms_train"]["rotation_range"],
                            width_shift_range=system_dict["local"]["transforms_train"]["width_shift_range"],
                            height_shift_range=system_dict["local"]["transforms_train"]["height_shift_range"],
                            shear_range=system_dict["local"]["transforms_train"]["shear_range"],
                            zoom_range=system_dict["local"]["transforms_train"]["zoom_range"],
                            brightness_range=system_dict["local"]["transforms_train"]["brightness_range"],
                            horizontal_flip=system_dict["local"]["transforms_train"]["horizontal_flip"],
                            vertical_flip=system_dict["local"]["transforms_train"]["vertical_flip"],
                            validation_split=0.9,
                            rescale=0
                        );

    if(system_dict["local"]["transforms_train"]["featurewise_center"]):
        system_dict["local"]["data_generators"]["estimate"].mean = system_dict["local"]["transforms_train"]["mean"];
    if(system_dict["local"]["transforms_train"]["featurewise_std_normalization"]):
        system_dict["local"]["data_generators"]["estimate"].std = system_dict["local"]["transforms_train"]["std"];

    return system_dict;




@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transform_trainval(system_dict):
    '''
    Load testing transforms in main state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    
    if("train-val" in system_dict["dataset"]["dataset_type"]):
        if(system_dict["local"]["transforms_train"]["featurewise_center"]):
            rescale_train = 1/256;
        else:
            rescale_train = 0;
        if(system_dict["local"]["transforms_val"]["featurewise_center"]):
            rescale_val = 1/256;
        else:
            rescale_val = 0;


        system_dict["local"]["data_generators"]["train"] = keras.preprocessing.image.ImageDataGenerator(
                            featurewise_center=system_dict["local"]["transforms_train"]["featurewise_center"],
                            featurewise_std_normalization=system_dict["local"]["transforms_train"]["featurewise_std_normalization"],
                            rotation_range=system_dict["local"]["transforms_train"]["rotation_range"],
                            width_shift_range=system_dict["local"]["transforms_train"]["width_shift_range"],
                            height_shift_range=system_dict["local"]["transforms_train"]["height_shift_range"],
                            shear_range=system_dict["local"]["transforms_train"]["shear_range"],
                            zoom_range=system_dict["local"]["transforms_train"]["zoom_range"],
                            brightness_range=system_dict["local"]["transforms_train"]["brightness_range"],
                            horizontal_flip=system_dict["local"]["transforms_train"]["horizontal_flip"],
                            vertical_flip=system_dict["local"]["transforms_train"]["vertical_flip"],
                            rescale = 0
                            );
                                       

       



        system_dict["local"]["data_generators"]["val"] = keras.preprocessing.image.ImageDataGenerator(
                            featurewise_center=system_dict["local"]["transforms_val"]["featurewise_center"],
                            featurewise_std_normalization=system_dict["local"]["transforms_val"]["featurewise_std_normalization"],
                            rotation_range=system_dict["local"]["transforms_val"]["rotation_range"],
                            width_shift_range=system_dict["local"]["transforms_val"]["width_shift_range"],
                            height_shift_range=system_dict["local"]["transforms_val"]["height_shift_range"],
                            shear_range=system_dict["local"]["transforms_val"]["shear_range"],
                            zoom_range=system_dict["local"]["transforms_val"]["zoom_range"],
                            brightness_range=system_dict["local"]["transforms_val"]["brightness_range"],
                            horizontal_flip=system_dict["local"]["transforms_val"]["horizontal_flip"],
                            vertical_flip=system_dict["local"]["transforms_val"]["vertical_flip"],
                            rescale = 0
                        );


        if(system_dict["local"]["transforms_train"]["featurewise_center"]):
            system_dict["local"]["data_generators"]["train"].mean = system_dict["local"]["transforms_train"]["mean"];
        if(system_dict["local"]["transforms_val"]["featurewise_center"]):
            system_dict["local"]["data_generators"]["val"].mean = system_dict["local"]["transforms_val"]["mean"];
        if(system_dict["local"]["transforms_train"]["featurewise_std_normalization"]):
            system_dict["local"]["data_generators"]["train"].std = system_dict["local"]["transforms_train"]["std"];
        if(system_dict["local"]["transforms_val"]["featurewise_std_normalization"]):
            system_dict["local"]["data_generators"]["val"].std = system_dict["local"]["transforms_val"]["std"];




    else:
        if(system_dict["local"]["transforms_train"]["featurewise_center"]):
            rescale_train = 1/256;
        else:
            rescale_train = 0;


        system_dict["local"]["data_generators"]["train"] = keras.preprocessing.image.ImageDataGenerator(
                            featurewise_center=system_dict["local"]["transforms_train"]["featurewise_center"],
                            featurewise_std_normalization=system_dict["local"]["transforms_train"]["featurewise_std_normalization"],
                            rotation_range=system_dict["local"]["transforms_train"]["rotation_range"],
                            width_shift_range=system_dict["local"]["transforms_train"]["width_shift_range"],
                            height_shift_range=system_dict["local"]["transforms_train"]["height_shift_range"],
                            shear_range=system_dict["local"]["transforms_train"]["shear_range"],
                            zoom_range=system_dict["local"]["transforms_train"]["zoom_range"],
                            brightness_range=system_dict["local"]["transforms_train"]["brightness_range"],
                            horizontal_flip=system_dict["local"]["transforms_train"]["horizontal_flip"],
                            vertical_flip=system_dict["local"]["transforms_train"]["vertical_flip"],
                            validation_split=1-system_dict["dataset"]["params"]["train_val_split"],
                            rescale = 0
                            );


        if(system_dict["local"]["transforms_train"]["featurewise_center"]):
            system_dict["local"]["data_generators"]["train"].mean = system_dict["local"]["transforms_train"]["mean"];
        if(system_dict["local"]["transforms_train"]["featurewise_std_normalization"]):
            system_dict["local"]["data_generators"]["train"].std = system_dict["local"]["transforms_train"]["std"];




    return system_dict;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transform_test(system_dict):
    if(system_dict["local"]["transforms_test"]["featurewise_center"]):
        rescale_val = 1/256;
    else:
        rescale_val = 0;

    system_dict["local"]["data_generators"]["test"] = keras.preprocessing.image.ImageDataGenerator(
                            featurewise_center=system_dict["local"]["transforms_test"]["featurewise_center"],
                            featurewise_std_normalization=system_dict["local"]["transforms_test"]["featurewise_std_normalization"],
                            rotation_range=system_dict["local"]["transforms_test"]["rotation_range"],
                            width_shift_range=system_dict["local"]["transforms_test"]["width_shift_range"],
                            height_shift_range=system_dict["local"]["transforms_test"]["height_shift_range"],
                            shear_range=system_dict["local"]["transforms_test"]["shear_range"],
                            zoom_range=system_dict["local"]["transforms_test"]["zoom_range"],
                            brightness_range=system_dict["local"]["transforms_test"]["brightness_range"],
                            horizontal_flip=system_dict["local"]["transforms_test"]["horizontal_flip"],
                            vertical_flip=system_dict["local"]["transforms_test"]["vertical_flip"],
                            rescale = 0
                            );

    if(system_dict["local"]["transforms_test"]["featurewise_center"]):
        system_dict["local"]["data_generators"]["test"].mean = system_dict["local"]["transforms_test"]["mean"];
    if(system_dict["local"]["transforms_test"]["featurewise_std_normalization"]):
        system_dict["local"]["data_generators"]["test"].std = system_dict["local"]["transforms_test"]["std"];

    return system_dict;


from monk.system.eda.imports import *
from monk.system.imports import *


@accepts(str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def get_img_label(fname, delimiter):
    '''
    Find list of images and corresponding labels in csv file

    Args:
        fname (str): Path to csv file
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of images in dataset
        list: List of labels corresponding every image in dataset
    '''
    f = open(fname);
    lst = f.readlines();
    f.close();
    del lst[0]
    img_list = [];
    label_list = [];
    for i in range(len(lst)):
        img, label = lst[i][:len(lst[i])-1].split(delimiter);
        img_list.append(img);
        label_list.append(label);

    return img_list, label_list;



@accepts(str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def read_csv(fname):
    '''
    Read CSV File

    Args:
        fname (str): Path to csv file

    Returns:
        list: List of every row in csv
    '''
    f = open(fname);
    lst = f.readlines();
    f.close();
    del lst[0]
    return lst;




@accepts(str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_from_folder_train(tpath):
    '''
    Find number of images in every class image folder - train

    Args:
        tpath (str): Path to image training folder

    Returns:
        list: List of classes
        list: List of number of images in every class
    '''
    classes_folder = os.listdir(tpath);
    classes_folder_strength = [];
    for i in range(len(classes_folder)):
        classes_folder_strength.append(len(os.listdir(tpath + "/" + classes_folder[i])));
    return classes_folder, classes_folder_strength;




@accepts(str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_from_folder_trainval(tpath, vpath):
    '''
    Find number of images in every class image folder - train and val

    Args:
        tpath (str): Path to image training folder
        vpath (str): Path to image validation folder

    Returns:
        list: List of classes
        list: List of number of images in every class
    '''
    classes_folder, classes_folder_strength = populate_from_folder_train(tpath);
    for i in range(len(os.listdir(vpath))):
            classes_folder_strength[i] = classes_folder_strength[i] + len(os.listdir(vpath + "/" + classes_folder[i]));
    return classes_folder, classes_folder_strength;


@accepts(str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_from_csv_train(tpath, delimiter):
    '''
    Find number of images in every class image csv - train

    Args:
        tpath (str): Path to csv pointing to training dataset
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of classes
        list: List of number of images in every class
    '''
    img_list, label_list = get_img_label(tpath, delimiter);
    classes_folder = list(np.unique(sorted(label_list)))
    classes_folder_strength = [];
    for i in range(len(classes_folder)):
        classes_folder_strength.append(label_list.count(classes_folder[i]));
    return classes_folder, classes_folder_strength;

@accepts(str, str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_from_csv_trainval(tpath, vpath, delimiter):
    '''
    Find number of images in every class image csv - train and val

    Args:
        tpath (str): Path to csv pointing to training dataset
        vpath (str): Path to csv pointing to validation dataset
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of classes
        list: List of number of images in every class
    '''
    classes_folder, classes_folder_strength = populate_from_csv_train(tpath, delimiter);
    img_list, label_list = get_img_label(vpath, delimiter);
    for i in range(len(classes_folder)):
        classes_folder_strength[i] += label_list.count(classes_folder[i]);
    return classes_folder, classes_folder_strength;


@accepts(str, str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_missing(tpath, dataset_path, delimiter):
    '''
    Find number of missing images in imageset

    Args:
        tpath (str): Path to csv pointing to training dataset
        dataset_path (str): Path to dataset containing images
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of missing images
    '''
    lst = read_csv(tpath);
    missing_images = [];
    for i in range(len(lst)):
        img, label = lst[i][:len(lst[i])-1].split(delimiter);
        if(not os.path.isfile(dataset_path + "/" + img)):
            missing_images.append(dataset_path + "/" + img);
    return missing_images;


@accepts(str, verbose=[bool, str, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_corrupt_from_foldered(dataset_path, verbose=1):
    '''
    Find number of corrupted images in imageset - foldered type

    Args:
        dataset_path (str): Path to dataset containing images
        verbose (str): If True, prints logs for analysis 

    Returns:
        list: List of corrupt images
    '''
    classes_folder = os.listdir(dataset_path);
    corrupt_images = [];
    if(verbose):
        for i in tqdm(range(len(classes_folder))):
            list_imgs = os.listdir(dataset_path + "/" + classes_folder[i]);  
            for j in range(len(list_imgs)):
                img_name = dataset_path + "/" + classes_folder[i] + "/" + list_imgs[j]
                if(os.path.isfile(img_name)):
                    try:
                        img = Image.open(img_name)
                        img.verify()
                    except Exception:
                        corrupt_images.append(img_name)
    else:
        for i in range(len(classes_folder)):
            list_imgs = os.listdir(dataset_path + "/" + classes_folder[i]);  
            for j in range(len(list_imgs)):
                img_name = dataset_path + "/" + classes_folder[i] + "/" + list_imgs[j]
                if(os.path.isfile(img_name)):
                    try:
                        img = Image.open(img_name)
                        img.verify()
                    except Exception:
                        corrupt_images.append(img_name)
    return corrupt_images;


@accepts(str, str, str, verbose=[bool, str, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def populate_corrupt_from_csv(tpath, dataset_path, delimiter, verbose=1):
    '''
    Find number of corrupted images in imageset - csv type

    Args:
        dataset_path (str): Path to dataset containing images
        verbose (str): If True, prints logs for analysis 

    Returns:
        list: List of corrupt images
    '''
    lst = read_csv(tpath);
    corrupt_images = [];
    if(verbose):
        for i in tqdm(range(len(lst))):
            img_name, label = lst[i][:len(lst[i])-1].split(delimiter);
            img_name = dataset_path + "/" + img_name
            if(os.path.isfile(img_name)):
                try:
                    img = Image.open(img_name)
                    img.verify()
                except Exception:
                    corrupt_images.append(img_name)
    else:
        for i in range(len(lst)):
            img_name, label = lst[i][:len(lst[i])-1].split(delimiter);
            img_name = dataset_path + "/" + img_name
            if(os.path.isfile(img_name)):
                try:
                    img = Image.open(img_name)
                    img.verify()
                except Exception:
                    corrupt_images.append(img_name)
    return corrupt_images;


from __future__ import print_function
from __future__ import division

import warnings
import mxnet as mx
import GPUtil
from mxnet import gluon, init, nd, initializer
from mxnet.gluon import nn
from gluoncv.model_zoo import get_model
import networkx as nx
from matplotlib import pyplot as plt

from mxnet.gluon.contrib import nn as contrib_nn

import os
import ipywidgets as widgets
from ipywidgets import interact, fixed
from IPython.display import display, clear_output
import matplotlib.gridspec as gridspec

ctx = [mx.cpu()];


layer_names = ["convolution1d", "convolution2d", "convolution", "convolution3d", "transposed_convolution1d",
                    "transposed_convolution", "transposed_convolution2d", "transposed_convolution3d", 
                    "max_pooling1d", "max_pooling2d", "max_pooling", "max_pooling3d", "average_pooling1d",
                    "average_pooling2d", "average_pooling", "average_pooling3d", "global_max_pooling1d",
                    "global_max_pooling2d", "global_max_pooling", "global_max_pooling3d", "global_average_pooling1d",
                    "global_average_pooling2d", "global_average_pooling", "global_average_pooling3d", 
                    "fully_connected", "dropout", "flatten", "identity", "add", "concatenate", "batch_normalization",
                    "instance_normalization", "layer_normalization", "relu", "sigmoid", "tanh", "softplus", "softsign", "elu", "gelu", "leaky_relu",
                    "prelu", "selu", "swish"]


names = ["conv1d_", "conv_", "conv_", "conv3d_", "tconv1d_", "tconv_", "tconv2d_", "tconv3d_", 
            "max-pool1d_", "mpool_", "mpool_", "max-pool3d_", "avg-pool1d_",
            "apool_", "apool_", "avg-pool3d_", "global-max-pool1d_", "global-max-pool2d_", "global-max-pool_", 
            "global-max-pool3d_", "global-avg-pool1d_", "global-avg-pool2d_", "global-avg-pool_", "global-avg-pool3d_", 
            "fc_", "dropout_", "flatten_", "identity_", "add_", "concat_", "bn_",
            "inorm_", "lnorm_", "relu_", "sigmoid_", "tanh_", "softplus_", "softsign_", "elu_", "gelu_", 
            "leaky_relu_", "prelu_", "selu_", "swish_"];








from monk.pytorch.finetune.imports import *
from monk.system.imports import *

from monk.pytorch.finetune.level_9_transforms_main import prototype_transforms


class prototype_schedulers(prototype_transforms):
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_fixed(self):
        '''
        Set learning rate fixed

        Args:
            None

        Returns:
            None
        '''
        self.system_dict = scheduler_fixed(self.system_dict);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, None, gamma=["gt", 0.01, "lt", 1], last_epoch=None, post_trace=False)
    @error_checks(None, ["gt", 0], gamma=["gt", 0], last_epoch=None, post_trace=False)
    @accepts("self", int, gamma=float, last_epoch=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_step_decrease(self, step_size, gamma=0.1, last_epoch=-1):
        '''
        Set learning rate to decrease in regular steps

        Args:
            step_size (int): Step interval for decreasing learning rate
            gamma (str): Reduction multiplier for reducing learning rate post every step
            last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

        Returns:
            None
        '''
        self.system_dict = scheduler_step(self.system_dict, step_size, gamma=gamma, last_epoch=last_epoch);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, None, gamma=["gt", 0.01, "lt", 1], last_epoch=None, post_trace=False)
    @error_checks(None, ["inc", None], gamma=["gt", 0], last_epoch=None, post_trace=False)
    @accepts("self", [list, int], gamma=float, last_epoch=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_multistep_decrease(self, milestones, gamma=0.1, last_epoch=-1):
        '''
        Set learning rate to decrease in irregular steps

        Args:
            milestones (list): List of epochs at which learning rate is to be decreased
            gamma (str): Reduction multiplier for reducing learning rate post every step
            last_epoch (int): Dummy variable

        Returns:
            None
        '''
        self.system_dict = scheduler_multistep(self.system_dict, milestones, gamma=gamma, last_epoch=last_epoch);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################





    ###############################################################################################################################################
    @warning_checks(None, ["gt", 0.01, "lt", 1], last_epoch=None, post_trace=False)
    @error_checks(None, ["gt", 0], last_epoch=None, post_trace=False)
    @accepts("self", [float, int], last_epoch=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_exponential_decrease(self, gamma, last_epoch=-1):
        '''
        Set learning rate to decrease exponentially every step

        Args:
            gamma (str): Reduction multiplier for reducing learning rate post every step
            last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

        Returns:
            None
        '''
        self.system_dict = scheduler_exponential(self.system_dict, gamma, last_epoch=last_epoch);

        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################






    ###############################################################################################################################################
    @warning_checks(None, mode=None, factor=["gt", 0.01, "lt", 1], patience=["lt", 20], verbose=None, threshold=None,
        threshold_mode=None, cooldown=None, min_lr=None, epsilon=["lt", 0.0001], post_trace=False)
    @error_checks(None, mode=["in", ["min", "max"]], factor=["gt", 0], patience=["gt", 0], verbose=None, threshold=["gte", 0], 
        threshold_mode=["in", ["rel", "abs"]], cooldown=["gte", 0], min_lr=["gte", 0], epsilon=["gte", 0], post_trace=False)
    @accepts("self", mode=str, factor=[float, int], patience=int, verbose=bool, threshold=[float, int], 
        threshold_mode=str, cooldown=int, min_lr=[float, list, int], epsilon=float, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_plateau_decrease(self, mode='min', factor=0.1, patience=10, verbose=False, threshold=0.0001, \
        threshold_mode='rel', cooldown=0, min_lr=0, epsilon=1e-08):
        '''
        Set learning rate to decrease if a metric (loss) stagnates in a plateau

        Args:
            mode (str): Either of 
                        - 'min' : lr will be reduced when the quantity monitored (loss) has stopped decreasing; 
                        - 'max' : lr reduced when the quantity monitored (accuracy) has stopped increasing. 
            factor (float): Reduction multiplier for reducing learning rate post every step
            patience (int): Number of epochs to wait before reducing learning rate
            verbose (bool): If True, all computations and wait times are printed
            threshold (float): Preset fixed to 0.0001
            threshold_mode (str): Preset fixed to 'rel' mode
            cooldown (int): Number of epochs to wait before actually applying the scheduler post the actual designated step
            min_lr (float): Set minimum learning rate, post which it will not be decreased
            epsilon (float): A small value to avoid divison by zero.
            last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

        Returns:
            None
        '''
        self.system_dict = scheduler_plateau(self.system_dict, mode=mode, factor=factor, patience=patience, verbose=verbose,
            threshold=threshold, threshold_mode=threshold_mode, cooldown=cooldown, min_lr=min_lr, epsilon=epsilon);

        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



from monk.gluon.datasets.imports import *
from monk.system.imports import *

@accepts(list, int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def balance_class_weights(label_list, nclasses):                        
    count = [0] * nclasses
    pbar=tqdm(total=len(label_list));
    for idx, val in enumerate(label_list):       
        pbar.update();
        count[val] += 1                                                     
    weight_per_class = [0.] * nclasses                                      
    N = float(sum(count))                                                   
    for i in range(nclasses):                                                   
        weight_per_class[i] = N/float(count[i])                                 
    weight = [0] * len(label_list)   
    pbar=tqdm(total=len(label_list));
    for idx, val in enumerate(label_list):  
        pbar.update();
        weight[idx] = weight_per_class[val]                                 
    return weight, weight_per_class, count;


from monk.pytorch.finetune.imports import *
from monk.system.imports import *

from monk.pytorch.finetune.level_12_losses_main import prototype_losses


class prototype_updates(prototype_losses):
    '''
    Main class for all parametric update functions

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ##########################################################################################################################################################
    @warning_checks(None, ["gte", 32, "lte", 1024], post_trace=False)
    @error_checks(None, ["gt", 0], post_trace=False)
    @accepts("self", int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_input_size(self, input_size):
        '''
        Update input size.

        Args:
            input_size (int): New input size

        Returns:
            None
        '''
        self.system_dict = set_input_size(input_size, self.system_dict);
        
        self.custom_print("Update: Input size - {}".format(self.system_dict["dataset"]["params"]["input_size"]));
        self.custom_print("");
        

    @warning_checks(None, ["lte", 128], post_trace=False)
    @error_checks(None, ["gt", 0], post_trace=False)
    @accepts("self", int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_batch_size(self, batch_size):
        '''
        Update batch size.

        Args:
            batch_size (int): New batch size

        Returns:
            None
        '''
        self.system_dict = set_batch_size(batch_size, self.system_dict);
        
        self.custom_print("Update: Batch size - {}".format(self.system_dict["dataset"]["params"]["batch_size"]));
        self.custom_print("");
        

    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_shuffle_data(self, shuffle):
        '''
        Update to shuffle data or not.

        Args:
            shuffle (bool): If True, will shuffle data

        Returns:
            None
        '''
        self.system_dict = set_data_shuffle(shuffle, self.system_dict);
        
        self.custom_print("Update: Data shuffle - {}".format(self.system_dict["dataset"]["params"]["train_shuffle"]));
        self.custom_print("");
        
    
    @warning_checks(None, ["lte", psutil.cpu_count()], post_trace=False)
    @error_checks(None, ["gt", 0], post_trace=False)
    @accepts("self", int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True) 
    def update_num_processors(self, num_processors):
        '''
        Update num processors for data loader.

        Args:
            num_processors (int): Max CPUs for data sampling

        Returns:
            None
        '''
        self.system_dict = set_num_processors(num_processors, self.system_dict);
        
        self.custom_print("Update: Num processors - {}".format(self.system_dict["dataset"]["params"]["num_workers"]));
        self.custom_print("");
        

    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_weighted_sampling(self, sample):
        '''
        Function inactive
        '''
        self.system_dict = set_weighted_sampling(sample, self.system_dict);
        
        self.custom_print("Update: Weighted Sampling - {}".format(self.system_dict["dataset"]["params"]["weighted_sample"]));
        self.custom_print("");



    @warning_checks(None, ["gt", 0.5, "lt", 1], post_trace=False)
    @error_checks(None, ["gt", 0, "lt", 1], post_trace=False)
    @accepts("self", float, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_trainval_split(self, value):
        '''
        Update training-validation split
        Args:
            split (float): Indicating train validation split
                            Division happens as follows:
                                train - total dataset * split * 100
                                val   - total dataset * (1-split) * 100 

        Returns:
            None
        '''
        if(self.system_dict["dataset"]["dataset_type"] == "train"):
            dataset_path = self.system_dict["dataset"]["train_path"];
            path_to_csv=False;
        elif(self.system_dict["dataset"]["dataset_type"] == "train-val"):
            dataset_path = [self.system_dict["dataset"]["train_path"], self.system_dict["dataset"]["val_path"]];
            path_to_csv=False;
        elif(self.system_dict["dataset"]["dataset_type"] == "csv_train"):
            dataset_path = self.system_dict["dataset"]["train_path"];
            path_to_csv = self.system_dict["dataset"]["csv_train"];
        elif(self.system_dict["dataset"]["dataset_type"] == "csv_train-val"):
            dataset_path = [self.system_dict["dataset"]["train_path"], self.system_dict["dataset"]["val_path"]];
            path_to_csv = [self.system_dict["dataset"]["csv_train"], self.system_dict["dataset"]["csv_val"]];
        else:
            msg = "Dataset Type invalid.\n";
            msg += "Cannot update split"
            ConstraintsWarning(msg)

        self.system_dict = set_dataset_train_path(self.system_dict, dataset_path, value, path_to_csv, self.system_dict["dataset"]["params"]["delimiter"]);


    @warning_checks(None, dataset_path=None, split=["gt", 0.5, "lt", 1], path_to_csv=None, delimiter=None, post_trace=False)
    @error_checks(None, dataset_path=["folder", 'r'], split=["gt", 0, "lt", 1], path_to_csv=["file", 'r'], delimiter=["in", [",", ";", "-", " "]], post_trace=False)
    @accepts("self", dataset_path=[str, list], split=float, path_to_csv=[str, list, bool], delimiter=str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_dataset(self, dataset_path=False, split=0.9, path_to_csv=False, delimiter=","):
        '''
        Update dataset path

        Args:
            dataset_path (str, list): Path to Dataset folder
                                      1) Single string if validation data does not exist
                                      2) List [train_path, val_path] in case of separate train and val data
            path_to_csv (str, list): Path to csv file pointing towards images
                                     1) Single string if validation data does not exist
                                     2) List [train_path, val_path] in case of separate train and val data
            value (float): Indicating train validation split
                            Division happens as follows:
                                train - total dataset * split * 100
                                val   - total dataset * (1-split) * 100 
            delimiter (str): Delimiter for csv file

        Returns:
            None
        '''
        self.system_dict = set_dataset_train_path(self.system_dict, dataset_path, split, path_to_csv, delimiter);
    ##########################################################################################################################################################




    ##########################################################################################################################################################
    @accepts("self", str, force=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_model_name(self, model_name, force=False):
        '''
        Update model name

        Args:
            model_name (str): Select from available models. Check via List_Models() function
            force (bool): Dummy function

        Returns:
            None
        '''
        if(not force):
            if(self.system_dict["training"]["status"]):
                ConstraintWarning("Model trained using {}\n".format(self.system_dict["model"]["params"]["model_name"]));
                ConstraintWarning("Changing the model will overwrite previously trained models if training is executed.\n");
                inp = input("Do you wish to continue further (y/n):");

                if(inp == "y"):
                    self.system_dict = set_model_name(model_name, self.system_dict);
                    self.custom_print("Update: Model name - {}".format(self.system_dict["model"]["params"]["model_name"]));
                    self.custom_print("");
                else:
                    self.custom_print("Model not updated.");
                    self.custom_print("");
            else:
                self.system_dict = set_model_name(model_name, self.system_dict);
                self.custom_print("Update: Model name - {}".format(self.system_dict["model"]["params"]["model_name"]));
                self.custom_print("");
        else:
            self.system_dict = set_model_name(model_name, self.system_dict);
            self.custom_print("Update: Model name - {}".format(self.system_dict["model"]["params"]["model_name"]));
            self.custom_print("");
    ##########################################################################################################################################################



    ##########################################################################################################################################################
    @accepts("self", [str, list], force=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_model_path(self, model_path, force=False):
        '''
        Update model path for inferencing

        Args:
            model_path (str): Path to model weights.
            force (bool): Dummy function

        Returns:
            None
        '''
        if(not force):
            if(self.system_dict["training"]["status"]):
                ConstraintWarning("Model trained using {}\n".format(self.system_dict["model"]["params"]["model_name"]));
                ConstraintWarning("Changing the model will overwrite previously trained models if training is executed.\n");
                inp = input("Do you wish to continue further (y/n):");

                if(inp == "y"):
                    self.system_dict = set_model_path(model_path, self.system_dict);
                    self.custom_print("Update: Model path - {}".format(self.system_dict["model"]["params"]["model_path"]));
                    self.custom_print("");
                else:
                    self.custom_print("Model not updated.");
                    self.custom_print("");
            else:
                self.system_dict = set_model_path(model_path, self.system_dict);
                self.custom_print("Update: Model path - {}".format(self.system_dict["model"]["params"]["model_path"]));
                self.custom_print("");
        else:
            self.system_dict = set_model_path(model_path, self.system_dict);
            self.custom_print("Update: Model path - {}".format(self.system_dict["model"]["params"]["model_path"]));
            self.custom_print("");
    ##########################################################################################################################################################
        


    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_use_gpu(self, gpu):
        '''
        Update to use gpu or cpu

        Args:
            gpu (bool): If True, then use GPU

        Returns:
            None
        '''
        self.system_dict = set_device(gpu, self.system_dict);
        
        self.custom_print("Update: Use Gpu - {}".format(self.system_dict["model"]["params"]["use_gpu"]));
        self.custom_print("");
    ##########################################################################################################################################################
        


    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_use_pretrained(self, pretrained):
        '''
        Update to use pretrained wights or randomly initialized weights

        Args:
            pretrained (bool): If True, use pretrained weights
                                else, use randomly initialized weights

        Returns:
            None
        '''
        self.system_dict = set_pretrained(pretrained, self.system_dict);
        
        self.custom_print("Update: Use pretrained - {}".format(self.system_dict["model"]["params"]["use_pretrained"]));
        self.custom_print("");
    ##########################################################################################################################################################
        


    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_freeze_base_network(self, freeze):
        '''
        Update whether freeze base network or not

        Args:
            freeze (bool): If True, then base network is non-trainable, works as a feature extractor

        Returns:
            None
        '''
        self.system_dict = set_freeze_base_network(freeze, self.system_dict);
        
        self.custom_print("Update: Freeze Base Network - {}".format(self.system_dict["model"]["params"]["freeze_base_network"]));
        self.custom_print("");
    ##########################################################################################################################################################


    ##########################################################################################################################################################
    @error_checks(None, ["gte", 0], post_trace=False)
    @accepts("self", int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_freeze_layers(self, num_freeze):
        '''
        Update to freeze certain layers in the network

        Args:
            num_freeze (int): Number of layers to freeze in network starting from top

        Returns:
            None
        '''
        self.system_dict["model"]["params"]["num_freeze"] = num_freeze;
        
        self.custom_print("Update: Freeze layers - {}".format(self.system_dict["model"]["params"]["num_freeze"]));
        self.custom_print("");
    ##########################################################################################################################################################




    ##########################################################################################################################################################
    @warning_checks(None, ["lt", 100], post_trace=False)
    @error_checks(None, ["gt", 0], post_trace=False)
    @accepts("self", int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_num_epochs(self, num_epochs):
        '''
        Update number of epochs to train the network

        Args:
            num_epochs (int): New number of epochs

        Returns:
            None
        '''
        self.system_dict = set_num_epochs(num_epochs, self.system_dict);
        
        self.custom_print("Update: Num Epochs - {}".format(self.system_dict["hyper-parameters"]["num_epochs"]));
        self.custom_print("");
    ##########################################################################################################################################################


    ##########################################################################################################################################################
    @warning_checks(None, ["lt", 1], post_trace=False)
    @error_checks(None, ["gt", 0], post_trace=False)
    @accepts("self", [int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_learning_rate(self, learning_rate):
        '''
        Update base learning rate for training

        Args:
            learning_rate (float): New base learning rate
            
        Returns:
            None
        '''
        self.system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
        self.system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;

        self.custom_print("Update: Learning Rate - {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("");
    ##########################################################################################################################################################



    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_display_progress_realtime(self, value):    
        '''
        Update display progress param

        Args:
            value (bool): If True, then real time progress is displayed
            
        Returns:
            None
        ''' 
        self.system_dict = set_display_progress_realtime(value, self.system_dict);
        
        self.custom_print("Update: Display progress realtime - {}".format(self.system_dict["training"]["settings"]["display_progress_realtime"]));
        self.custom_print("");
    ##########################################################################################################################################################
        

    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_display_progress(self, value): 
        '''
        Update display progress param

        Args:
            value (bool): If True, then per epoch progress is displayed
            
        Returns:
            None
        ''' 
        self.system_dict = set_display_progress(value, self.system_dict);
        
        self.custom_print("Update: Display progress  - {}".format(self.system_dict["training"]["settings"]["display_progress"]));
        self.custom_print("");
    ##########################################################################################################################################################
        

    ##########################################################################################################################################################
    @error_checks(None, None, prefix=["name", ["A-Z", "a-z", "0-9", "-", "_"]], post_trace=False)
    @accepts("self", bool, prefix=str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_save_intermediate_models(self, value, prefix="intermediate_model_"): 
        '''
        Update whether to save intermediate models or not

        Args:
            value (bool): If True, saves model weight post every epoch
            prefix (str): Appends a prefix to intermediate weights
            
        Returns:
            None
        '''
        if(value):
            if(not os.access(self.system_dict["model_dir"], os.W_OK)):
                msg = "Folder \"{}\" has no read access".format(self.system_dict["model_dir"])
                msg += "Cannot save Intermediate models";
                raise ConstraintError(msg);

        self.system_dict = set_save_intermediate_models(value, self.system_dict);
        self.system_dict = set_intermediate_model_prefix(prefix, self.system_dict);
        
        self.custom_print("Update: Save Intermediate models - {}".format(self.system_dict["training"]["settings"]["save_intermediate_models"]));
        if(self.system_dict["training"]["settings"]["save_intermediate_models"]):
            self.custom_print("Update: Intermediate model prefix - {}".format(self.system_dict["training"]["settings"]["intermediate_model_prefix"]));
        self.custom_print("");
    ##########################################################################################################################################################
        

    ##########################################################################################################################################################
    @accepts("self", bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def update_save_training_logs(self, value):
        '''
        Update whether to save training logs or not

        Args:
            value (bool): If True, saves all training and validation metrics. Required for comparison.
            
        Returns:
            None
        '''
        self.system_dict = set_save_training_logs(value, self.system_dict);
        
        self.custom_print("Update: Save Training logs - {}".format(self.system_dict["training"]["settings"]["save_training_logs"]));
        self.custom_print("");
    ##########################################################################################################################################################


import os
import sys

import psutil

from monk.pytorch_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.pytorch.common import print_start
from monk.pip_unit_tests.pytorch.common import print_status

import torch
import numpy as np
from monk.pytorch.losses.return_loss import load_loss


def test_activation_softsign(system_dict):
    forward = True;

    test = "test_activation_softsign";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");


            network = [];
            network.append(gtf.softsign());
            gtf.Compile_Network(network, data_shape=(3, 64, 64), use_gpu=False);

            x = torch.randn(1, 3, 64, 64);
            y = gtf.system_dict["local"]["model"](x);         

            system_dict["successful_tests"] += 1;
            print_status("Pass");

        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.tf_keras_1.models.imports import *
from monk.system.imports import *
from monk.tf_keras_1.models.layers import get_layer



@accepts("self", bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_parameter_requires_grad(finetune_net, freeze_base_network):
    '''
    Freeze based network as per params set

    Args:
        finetune_net (network): Model network
        freeze_base_network (bool): If True, all trainable params are freezed

    Returns:
        network: Updated Model network
    '''
    if freeze_base_network:
        for layer in finetune_net.layers:
            layer.trainable=False
    else:
        for layer in finetune_net.layers:
            layer.trainable=True

    return finetune_net



@accepts(list, "self", post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def set_final_layer(custom_network, x):
    '''
    Setup final sub-network 

    Args:
        custom_network (list): List of dicts containing details on appeded layers to base netwoek in transfer learning
        num_ftrs (int): Number of features coming from base network's last layers
        num_classes (int): Number of classes in the dataset

    Returns:
        layer: Sequential sub-network with added layers 
    '''
    for i in range(len(custom_network)):
        x = get_layer(custom_network[i], x);
    return x;





@accepts("self", list, int, set=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def create_final_layer(finetune_net, custom_network, num_classes):
    '''
    Create final sub-network 

    Args:
        finetune_net (network): Initial base network
        custom_network (list):  List of dicts containing details on appeded layers to base netwoek in transfer learning
        num_classes (int): Number of classes in the dataset

    Returns:
        network: Updated base network with appended custom additions
    '''
    x = finetune_net.output
    x = krl.GlobalAveragePooling2D()(x);

    preds = set_final_layer(custom_network, x);

    finetune_net = keras.models.Model(inputs=finetune_net.input, outputs=preds);
    return finetune_net;





@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_num_layers(system_dict):
    '''
    Depricated function - Get number of potentially trainable layers

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    complete_list = [];
    for layer in system_dict["local"]["model"].layers:
        if(layer.count_params() > 0):
            complete_list.append(layer.name)
    system_dict["model"]["params"]["num_layers"] = len(complete_list);
    return system_dict;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_num_trainable_layers(system_dict):
    '''
    Get number of potentially trainable layers

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    complete_list_trainable = [];
    for layer in system_dict["local"]["model"].layers:
        if(layer.count_params() > 0):
            if(layer.trainable):
                complete_list_trainable.append(layer.name)
    system_dict["model"]["params"]["num_params_to_update"] = len(complete_list_trainable);
    return system_dict;




@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def freeze_layers(num, system_dict):
    '''
    Main function responsible to freeze layers in network

    Args:
        num (int): Number of layers to freeze
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    system_dict = get_num_layers(system_dict);
    num_layers_in_model = system_dict["model"]["params"]["num_layers"];
    if(num > num_layers_in_model):
        msg = "Parameter num > num_layers_in_model\n";
        msg += "Freezing entire network\n";
        msg += "TIP: Total layers: {}".format(num_layers_in_model);
        raise ConstraintError(msg);

    current_num = 0;
    value = False;
    for layer in system_dict["local"]["model"].layers:
        if(layer.count_params() > 0):
            layer.trainable = value;
            current_num += 1;
            if(current_num == num):
                value = True;   

    system_dict = get_num_trainable_layers(system_dict);
    system_dict["model"]["status"] = True;   

    return system_dict;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def print_grad_stats(system_dict):
    '''
    Print details on which layers are trainable

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        None 
    '''
    print("Model - Gradient Statistics");
    i = 1;
    for layer in system_dict["local"]["model"].layers:
        print("    {}. {} Trainable - {}".format(i+1, layer.name, layer.trainable));
        i += 1;
    print("");





@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def setup_device_environment(system_dict):
    '''
    Load model weights on device - cpu or gpu 

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    num_cores = psutil.cpu_count();
    if system_dict["model"]["params"]["use_gpu"]:
        num_GPU = 1
        num_CPU = 1
    else:
        num_CPU = 1
        num_GPU = 0

    if(tf.__version__.split(".")[0] == "2"):
        gpu_options = tf.compat.v1.GPUOptions(per_process_gpu_memory_fraction=system_dict["model"]["params"]["gpu_memory_fraction"], 
            allow_growth = True)
        
        config = tf.compat.v1.ConfigProto(intra_op_parallelism_threads=num_cores, 
            inter_op_parallelism_threads=num_cores, allow_soft_placement=True, 
            device_count = {'CPU' : num_CPU, 'GPU' : num_GPU}, 
            gpu_options=gpu_options)
        
        session = tf.compat.v1.Session(config=config)
    else:
        gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=system_dict["model"]["params"]["gpu_memory_fraction"], 
            allow_growth = True)
        
        config = tf.ConfigProto(intra_op_parallelism_threads=num_cores, 
            inter_op_parallelism_threads=num_cores, allow_soft_placement=True, 
            device_count = {'CPU' : num_CPU, 'GPU' : num_GPU}, 
            gpu_options=gpu_options)
        
        session = tf.Session(config=config)

    K.set_session(session);
    
    return system_dict;




@accepts(dict, list, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_layer_uid(network_stack, count):
    '''
    Get a unique name for layer in custom network development

    Args:
        network_stack (list): List of list containing custom network details
        count (dict): a unique dictionary mapping number of every type of layer in the network
        system_dict (dict): System dict containing system state and parameters

    Returns:
        str: layer unique name
        dict: updated layer type mapper count
    '''
    if network_stack["uid"]:
        return network_stack["uid"], count;
    else:
        index = layer_names.index(network_stack["name"]);
        network_name = names[index] + str(count[index]);
        count[index] += 1;
        return network_name, count;
    





        
